;; Generated package description from gnosis.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "gnosis" "0.7.0" "Spaced Repetition System" '((emacs "27.2") (emacsql "4.1.0") (compat "29.1.4.2") (transient "0.7.2") (org-gnosis "0.0.9")) :commit "54660aabfa3200663053dd0063f2c215ef674ddb" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("extensions") :url "https://thanosapollo.org/projects/gnosis")
